from django import template
register = template.Library()


@register.filter(name='httpdate')
def httpdate(dt):
    """
    Convert a datetime object into a string in RFC 1123 format.
    """
    return dt.strftime('%a, %d %b %Y %H:%M:%S GMT')



import urllib
from django.core.urlresolvers import reverse
from django.http import HttpResponse, Http404
from dateutil.parser import parse as dateparser
from django.utils.cache import patch_vary_headers
from django.utils.translation import ugettext as _
from memento.templatetags.memento_tags import httpdate
from django.core.exceptions import ImproperlyConfigured
from django.contrib.syndication.views import add_domain
from django.contrib.sites.models import get_current_site
from django.views.generic import RedirectView, DetailView


class MementoDetailView(DetailView):
    """
    Extends Django's DetailView to describe an archived resource.

    A set of extra headers is added to the response. They are:

        * The "Memento-Datetime" when the resource was archived.

        * A "Link" that includes the URL of the original resource as well
          as the location of the TimeMap the publishes the directory of
          all versions archived by your site.

        * A "Vary" indicator that this is a Memento enabled site

    The view should be subclassed and configured like any other DetailView
    with a few extra options. They are:

        * datetime_field: A string attribute that is the name of the database
          field that contains the timestamp when the resource was archived.

        * timemap_pattern: The name of the URL pattern for this site's TimeMap
          that, given the original url, is able to reverse to return the
          location of the map that serves as the directory of all versions of
          this resource archived by your site.

        * get_original_url: A method that, given the object being rendered
          by the view, will return the original URL of the archived resource.

    """
    datetime_field = 'datetime'
    timemap_pattern_name = None

    def get_timemap_url(self, request, url):
        """
        Returns the location of the TimeMap that lists resources archived
        for the provided URL.
        """
        path = reverse(self.timemap_pattern_name, kwargs={'url': url})
        current_site = get_current_site(request)
        return add_domain(
            current_site.domain,
            path,
            request.is_secure(),
        )

    def get_original_url(self, obj):
        raise NotImplementedError("get_original_url method not implemented")

    def get(self, request, *args, **kwargs):
        response = super(MementoDetailView, self).get(
            request,
            *args,
            **kwargs
        )
        dt = getattr(self.object, self.datetime_field)
        response['Memento-Datetime'] = httpdate(dt)
        patch_vary_headers(response, ["accept-datetime"])
        if self.timemap_pattern_name:
            original_url = self.get_original_url(self.object)
            timemap_url = self.get_timemap_url(self.request, original_url)
            response['Link'] = """<%(original_url)s>; rel="original", \
<%(timemap_url)s>; rel="timemap"; type="application/link-format\"""" % dict(
                original_url=urllib.unquote(original_url),
                timemap_url=urllib.unquote(timemap_url)
            )
        return response


class TimeGateView(RedirectView):
    """
    Creates a TimeGate that handles a request with Memento headers
    and returns a response that redirects to the corresponding
    Memento.
    """
    model = None
    queryset = None
    timemap_pattern_name = None
    url_kwarg = 'url'
    url_field = 'url'
    datetime_field = 'datetime'

    def parse_datetime(self, request):
        """
        Parses the requested datetime from the request headers
        and returns it if it exists. Otherwise returns None.
        """
        # Verify that the Accept-Datetime header is provided
        if not request.META.get("HTTP_ACCEPT_DATETIME"):
            return None

        # Verify that the Accept-Datetime header is valid
        dt = dateparser(request.META.get("HTTP_ACCEPT_DATETIME"))
        if not dt:
            return HttpResponse(
                _("Bad request (400): Accept-Datetime header is malformed"),
                status=400
            )
        return dt

    def get_object(self, url, dt):
        """
        Accepts the requested URL and datetime and returns the object
        with the smallest date difference.
        """
        queryset = self.get_queryset()
        queryset = queryset.filter(**{self.url_field: url})

        try:
            prev_obj = queryset.filter(
                **{"%s__lte" % self.datetime_field: dt}
            ).order_by("-%s" % self.datetime_field)[0]
        except IndexError:
            raise Http404(_("No %(verbose_name)s found matching the query") %
                          {'verbose_name': queryset.model._meta.verbose_name})

        try:
            next_obj = queryset.filter(
                **{"%s__gte" % self.datetime_field: dt}
            ).order_by("%s" % self.datetime_field)[0]
        except IndexError:
            next_obj = None

        if not next_obj:
            return prev_obj
        else:
            prev_delta = abs(dt - getattr(prev_obj, self.datetime_field))
            next_delta = abs(dt - getattr(next_obj, self.datetime_field))
            if prev_delta <= next_delta:
                return prev_obj
            else:
                return next_obj

    def get_most_recent_object(self, url):
        """
        Returns the most recently archive object of the submitted URL
        """
        queryset = self.get_queryset()
        try:
            return queryset.filter(**{
                self.url_field: url,
            }).order_by("-%s" % self.datetime_field)[0]
        except IndexError:
            raise Http404(_("No %(verbose_name)s found matching the query") %
                          {'verbose_name': queryset.model._meta.verbose_name})

    def get_queryset(self):
        """
        Return the `QuerySet` that will be used to look up the object.
        Note that this method is called by the default implementation of
        `get_object` and may not be called if `get_object` is overridden.
        """
        if self.queryset is None:
            if self.model:
                return self.model._default_manager.all()
            else:
                raise ImproperlyConfigured(
                    "%(cls)s is missing a QuerySet. Define "
                    "%(cls)s.model, %(cls)s.queryset, or override "
                    "%(cls)s.get_queryset()." % {
                        'cls': self.__class__.__name__
                    }
                )
        return self.queryset.all()

    def get_redirect_url(self, request, obj):
        """
        Returns the URL that will redirect to the Memento resource.
        """
        current_site = get_current_site(request)
        return add_domain(
            current_site.domain,
            obj.get_absolute_url(),
            request.is_secure(),
        )

    def get_timemap_url(self, request, url):
        """
        Returns the location of the TimeMap that lists resources archived
        for the provided URL.
        """
        path = reverse(self.timemap_pattern_name, kwargs={'url': url})
        current_site = get_current_site(request)
        return add_domain(
            current_site.domain,
            path,
            request.is_secure(),
        )

    def get(self, request, *args, **kwargs):
        url = self.kwargs.get(self.url_kwarg)
        if not url:
            return HttpResponse(
                _("Bad request (400): URL not provided"),
                status=400
            )
        url = url.replace("http:/", "http://")
        url = url.replace("http:///", "http://")
        dt = self.parse_datetime(request)
        if dt:
            obj = self.get_object(url, dt)
        else:
            obj = self.get_most_recent_object(url)
        redirect_url = self.get_redirect_url(request, obj)
        response = HttpResponse(status=302)
        patch_vary_headers(response, ["accept-datetime"])
        if self.timemap_pattern_name:
            response['Link'] = """<%(url)s>; rel="original", \
<%(timemap_url)s>; rel="timemap"; type="application/link-format\"""" % dict(
                url=urllib.unquote(url),
                timemap_url=urllib.unquote(self.get_timemap_url(request, url))
            )
        response['Location'] = urllib.unquote(redirect_url)
        return response



from django.utils.six import StringIO
from django.utils.encoding import iri_to_uri
from django.template.loader import render_to_string


class TimemapLinkListGenerator(object):
    """
    Base class for creating a timemap.
    """
    mime_type = 'application/link-format; charset=utf-8'
    template_name = "memento/timemap/link_list.txt"

    def __init__(self, original_url, timemap_url):
        self.feed = {
            'original_url': iri_to_uri(original_url),
            'timemap_url': iri_to_uri(timemap_url),
        }
        self.items = []

    def add_item(self, link, datetime, first=False, last=False):
        """
        Adds an item to the feed.
        """
        item = {
            'link': iri_to_uri(link),
            'datetime': datetime,
            'first': first,
            'last': last,
        }
        self.items.append(item)

    def minimum_datetime(self):
        """
        Returns the earliest datetime in the item list.
        """
        return min([i['datetime'] for i in self.items])

    def maximum_datetime(self):
        """
        Returns the latest datetime in the item list.
        """
        return max([i['datetime'] for i in self.items])

    def get_context(self):
        return {
            'original_url': self.feed['original_url'],
            'timemap_url': self.feed['timemap_url'],
            'minimum_datetime': self.minimum_datetime(),
            'maximum_datetime': self.maximum_datetime(),
            'items': self.items,
        }

    def write(self, outfile, encoding):
        context = self.get_context()
        s = render_to_string(self.template_name, context)
        outfile.write(s)

    def writeString(self, encoding):
        """
        Returns the feed in the given encoding as a string.
        """
        s = StringIO()
        self.write(s, encoding)
        return s.getvalue()


class TimemapLinkIndexGenerator(TimemapLinkListGenerator):
    template_name = "memento/timemap/link_index.txt"

    def add_item(self, link, minimum_datetime=None, maximum_datetime=None):
        """
        Adds an item to the feed.
        """
        item = {
            'link': iri_to_uri(link),
            'minimum_datetime': minimum_datetime,
            'maximum_datetime': maximum_datetime,
        }
        self.items.append(item)

    def minimum_datetime(self):
        """
        Returns the earliest datetime in the item list.
        """
        return min([i['minimum_datetime'] for i in self.items])

    def maximum_datetime(self):
        """
        Returns the latest datetime in the item list.
        """
        return max([i['maximum_datetime'] for i in self.items])

    def get_context(self):
        return {
            'original_url': self.feed['original_url'],
            'timemap_url': self.feed['timemap_url'],
            'minimum_datetime': self.minimum_datetime(),
            'maximum_datetime': self.maximum_datetime(),
            'items': self.items,
        }



from django.utils import six
from django.templatetags.tz import utc
from django.utils.timezone import is_naive
from django.http import Http404, HttpResponse
from django.contrib.sites.models import get_current_site
from django.core.paginator import InvalidPage, Paginator
from django.contrib.syndication.views import add_domain
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from .feedgenerator import TimemapLinkListGenerator, TimemapLinkIndexGenerator


class TimemapLinkList(object):
    """
    A feed class that returns a list in Memento's Timemap format.
    """
    paginator_class = Paginator
    paginate_by = None
    page_kwarg = 'page'

    def __call__(self, request, *args, **kwargs):
        try:
            obj = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            raise Http404('Feed object does not exist.')
        self.request = request
        self.current_site = get_current_site(request)
        self.queryset = self.__get_dynamic_attr('memento_list', obj)
        feedgen = self.get_feed(obj)
        response = HttpResponse(content_type=feedgen.mime_type)
        feedgen.write(response, 'utf-8')
        return response

    def __get_dynamic_attr(self, attname, obj, default=None):
        try:
            attr = getattr(self, attname)
        except AttributeError:
            return default
        if callable(attr):
            try:
                code = six.get_function_code(attr)
            except AttributeError:
                code = six.get_function_code(attr.__call__)
            if code.co_argcount == 2:
                return attr(obj)
            else:
                return attr()
        return attr

    def memento_datetime(self, item):
        raise ImproperlyConfigured('Define an item_datetime() method in \
your %s class.' % self.__class__.__name__)

    def memento_link(self, memento):
        try:
            return memento.get_absolute_url()
        except AttributeError:
            raise ImproperlyConfigured(
                'Give your %s class a get_absolute_url() method, or define a '
                'memento() method in your TimemapLinkList class.' % (
                    memento.__class__.__name__
                )
            )

    def get_page_number(self):
        page = self.request.GET.get(self.page_kwarg) or None
        if not page:
            return None
        try:
            return int(page)
        except ValueError:
            raise Http404("Page can't be converted to an int.")

    def get_paginator(self, queryset):
        return self.paginator_class(queryset, self.paginate_by)

    def get_page(self, page_number):
        paginator = self.get_paginator(self.queryset)
        try:
            return paginator.page(page_number)
        except InvalidPage as e:
            raise Http404('Invalid page (%(page_number)s): %(message)s' % {
                'page_number': page_number,
                'message': str(e)
            })

    def get_list_feed(self, obj, page_number=None):
        feed_type = TimemapLinkListGenerator
        feed = feed_type(
            original_url=self.get_original_url(obj),
            timemap_url=add_domain(
                self.current_site.domain,
                self.request.path,
                self.request.is_secure(),
            ),
        )
        if page_number:
            page = self.get_page(page_number)
            self.queryset = page.object_list
        item_list = []
        for item in self.queryset:
            link = add_domain(
                self.current_site.domain,
                self.__get_dynamic_attr('memento_link', item),
                self.request.is_secure(),
            )
            item_datetime = self.__get_dynamic_attr('memento_datetime', item)
            if item_datetime and is_naive(item_datetime):
                item_datetime = utc(item_datetime)
            item_list.append(dict(
                link=link,
                datetime=item_datetime,
            ))
        if not page_number:
            item_list[0]['first'] = True
            item_list[-1]['last'] = True
        else:
            if page_number == 1:
                item_list[0]['first'] = True
            elif not page.has_next():
                item_list[-1]['last'] = True
        [feed.add_item(**d) for d in item_list]
        return feed

    def get_index_feed(self, obj):
        feed_type = TimemapLinkIndexGenerator
        timemap_url = add_domain(
            self.current_site.domain,
            self.request.path,
            self.request.is_secure(),
        )
        feed = feed_type(
            original_url=self.get_original_url(obj),
            timemap_url=timemap_url,
        )
        paginator = self.get_paginator(self.queryset)
        item_list = []
        for page in paginator.page_range:
            link = add_domain(
                self.current_site.domain,
                "%s?%s=%s" % (timemap_url, self.page_kwarg, page),
                self.request.is_secure(),
            )
            item_list.append(dict(link=link))
        [feed.add_item(**d) for d in item_list]
        return feed

    def get_feed(self, obj):
        """
        Returns a feedgenerator.DefaultFeed object, fully populated, for
        this feed. Raises FeedDoesNotExist for invalid parameters.
        """
        if self.paginate_by:
            page_number = self.get_page_number()
            if page_number:
                return self.get_list_feed(obj, page_number)
            else:
                return self.get_index_feed(obj)
        else:
            return self.get_list_feed(obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from distutils.core import Command


class TestCommand(Command):
    user_options = []
    package = 'memento'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from django.conf import settings
        settings.configure(
            DATABASES={
                'default': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3'
                }
            },
            INSTALLED_APPS=(self.package,),
            MIDDLEWARE_CLASSES=()
        )
        from django.core.management import call_command
        import django
        if django.VERSION[:2] >= (1, 7):
            django.setup()
        call_command('test', self.package)


setup(
    name='django-memento-framework',
    version='0.0.5',
    description='A set for helpers for Django web sites to enable the \
Memento framework for time-based access',
    author='Ben Welsh',
    author_email='ben.welsh@gmail.com',
    url='http://django-memento-framework.readthedocs.org',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,  # because we're including static files
    install_requires=(
        'python-dateutil',
    ),
    license="MIT",
    cmdclass={'test': TestCommand},
    classifiers=(
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ),
)